#!/usr/bin/env python3

# my own SAT/CNF library

# dennis(a)yurichev.com, 2017-2020

# "BV" stands for bitvector

# TODO: check signed/unsigned issues in adder/multiplier/divider
# TODO: instance class? term class?
# TODO: get rid of *fix* functions?

from ctypes import *
import subprocess, os, itertools, sys, my_utils
from typing import List
from typing import Tuple

# BV=[MSB...LSB]
def BV_to_number(BV:List[int]) -> int:
    # coeff=1, 2^1, 2^2 ... 2^(len(BV)-1)
    coeff=1
    rt=0
    for v in my_utils.rvr(BV):
        rt=rt+coeff*v
        coeff=coeff*2
    return rt

# bit order: [MSB..LSB]
# 'size' is desired width of bitvector, in bits:
def n_to_BV (n:int, size:int) -> List[bool]:
    out=[False]*size
    i=0
    for var in my_utils.rvr(list(bin(n)[2:])):
        if var=='1':
            out[i]=True
        else:
            out[i]=False
        i=i+1
    return my_utils.rvr(out)

def negate_clause(s:List[str]) -> List[str]:
    rt=[]
    for i in s:
        if i=="0":
            continue
        rt.append(i[1:] if i.startswith("-") else "-"+i)
    return rt

class picosat:

    def __init__(self):
        self.lib=None
        #print (sys.path)
        for d in sys.path:
            fname=d+"/libpicosat.so"
            if os.path.exists(fname):
                #print (fname)
                self.lib=CDLL(fname)
                break

        if self.lib==None:
            print ("Error: Can't find libpicosat.so")
            exit(0)

        self.lib.picosat_init.restype=c_void_p

        self.lib.picosat_reset.argtypes=[c_void_p]

        self.lib.picosat_add.argtypes=[c_void_p, c_int]
        self.lib.picosat_add.restype=c_int

        # int picosat_add_lits (PicoSAT *, int * lits);
        self.lib.picosat_add_lits.argtypes=[c_void_p, POINTER(c_int)]
        self.lib.picosat_add_lits.restype=c_int

        self.lib.picosat_assume.argtypes=[c_void_p, c_int]

        self.lib.picosat_sat.argtypes=[c_void_p, c_int]
        self.lib.picosat_sat.restype=c_int

        self.lib.picosat_deref.argtypes=[c_void_p, c_int]
        self.lib.picosat_deref.restype=c_int

        self.ctx=self.lib.picosat_init()

    def reset (self):
        self.lib.picosat_reset(self.ctx)

    def add (self, i):
        return self.lib.picosat_add(self.ctx, i)

    def add_lits (self, lits):
        arg2=(c_int * len(lits))(*lits)
        return self.lib.picosat_add_lits(self.ctx, arg2)

    def assume (self, i):
        self.lib.picosat_assume(self.ctx, i)

    def sat (self):
        return self.lib.picosat_sat(self.ctx, -1)

    def deref (self, i):
        return self.lib.picosat_deref(self.ctx, i)

"""
test:

o=picosat()
o.add(1)
o.add(0)
#o.add(-2)
#o.add(0)
print (o.sat())
o.assume(-2)
print (o.sat())
print ("calling deref")
print (o.deref(1))
print (o.deref(2))
"""

class SAT_lib:

    # SAT_solver can be: "minisat", "plingeling", "picosat", "libpicosat", "picomus"...
    def __init__(self, maxsat:bool=False, SAT_solver:str="minisat", verbose=0):
        self.SAT_solver=SAT_solver
        self.verbose=verbose
        self.last_var=1 # unused var...

        # just list of lines to be written to CNF-file:
        if SAT_solver != "libpicosat":
            self.CNF_next_idx=0
            self.CNF_clauses={} # k=int, v=List[int]
            self.CNF_comments={} # k=int, v=str
            self.CNF_initialized=True
        else:
            self.libpicosat=picosat()
            self.CNF_initialized=False

        self.maxsat=maxsat
        if self.maxsat==False:
            self.CNF_fname="1.cnf"
        else:
            self.CNF_fname="1.wcnf"

        self.clauses_total=0
        #self.HARD_CLAUSE=10000
        self.HARD_CLAUSE=50000000

        #self.remove_CNF_file=True
        self.remove_CNF_file=False

        self.verbosity=0

        # allocate a single variable fixed to False:
        self.const_false=self.create_var()
        self.fix(self.const_false, False)
        # allocate a single variable fixed to True:
        self.const_true=self.create_var()
        self.fix(self.const_true, True)

        self.OR_list_already_generated={}
        self.AND_already_generated={}
        self.AND_list_already_generated={}
        self.NAND_already_generated={}
        self.NOT_already_generated={}
        self.XOR_already_generated={}
        self.EQ_already_generated={}
        self.IMPLY_already_generated={}
        self.ITE_already_generated={}

    # only for libpicosat:
    def assume (self, v:int):
        assert self.SAT_solver=="libpicosat"
        self.libpicosat.assume(v)

    def run_minisat (self):
        try:
            child = subprocess.Popen(["minisat", self.CNF_fname, "results.txt"], stdout=subprocess.PIPE)
        except FileNotFoundError:
            print ("minisat not found. install it: 'sudo apt install minisat' or 'sudo pkg_add -i minisat'")
            exit(0)
        child.wait()
        # 10 is SAT, 20 is UNSAT
        if child.returncode==20:
            os.remove ("results.txt")
            return None

        if child.returncode!=10:
            print ("(minisat) unknown retcode: ", child.returncode)
            exit(0)

        #print "minisat done"
        t=my_utils.read_lines_from_file("results.txt")[1].split(" ")
        # remove last "variable", which is 0
        assert t[-1]=='0'
        t=t[:-1]
        # there was a time whan $t$ list was returned as solution!
        # now it's dict
        solution={}
        for i in t:
            if i.startswith("-"):
                solution[-int(i)]=False
            else:
                solution[int(i)]=True

        os.remove ("results.txt")

        return solution
    
    # cmd_line is an array
    def run_plingeling_or_open_wbo (self, cmd_line:str):
        tmp_fname="tmp.out"
        err_fname="tmp.err"
        logfile=open(tmp_fname, "w")
        errfile=open(err_fname, "w")
        try:
            child = subprocess.Popen(cmd_line, stdout=logfile, stderr=errfile)
        except FileNotFoundError:
            print ("can't run: "+" ".join(cmd_line))
            exit(0)

        child.wait()
        logfile.flush()
        logfile.close()
        errfile.flush()
        errfile.close()

        errfile=open(err_fname, "r")
        line = errfile.readline()
        if len(line)>0:
            print ("error log:", line)
            exit(0)
        errfile.close()

        tmp=[]
        logfile=open(tmp_fname, "r")
        while True:
            line = logfile.readline()
            #print ("line=", line)
            if line.startswith("s UNSAT"):
                logfile.close()
                os.remove(tmp_fname)
                os.remove(err_fname)
                return None
            elif line.startswith("v "):
                tmp.append (line[2:].rstrip())
            elif line=='':
                break
            else:
                pass

        logfile.close()
        os.remove(tmp_fname)
        os.remove(err_fname)
        t=" ".join(tmp).split(" ")
        # there was a time whan $t$ list was returned as solution!
        # now it's dict
        solution={}
        for i in t:
            if i.startswith("-"):
                solution[-int(i)]=False
            else:
                solution[int(i)]=True
        return solution

    def run_libpicosat(self):
        t=self.libpicosat.sat()
        if t==10: # SAT
            solution={}
            for v in range(1, self.last_var):
                x=self.libpicosat.deref(v)
                if x==-1:
                    solution[v]=False
                elif x==1:
                    solution[v]=True
                else:
                    raise ValueError ("unknown code returned by libpicosat.deref(): "+str(v))
            return solution
        elif t==20: # UNSAT
            return None
        else: 
            raise ValueError ("unknown code returned by libpicosat.sat(): "+str(t))

    def run_sat_solver(self):
        if self.SAT_solver=="minisat":
            return self.run_minisat()
        elif self.SAT_solver=="plingeling":
            return self.run_plingeling_or_open_wbo(["plingeling", self.CNF_fname])
        elif self.SAT_solver=="picosat":
            return self.run_plingeling_or_open_wbo(["picosat", self.CNF_fname])
        elif self.SAT_solver=="libpicosat":
            return self.run_libpicosat()
        else:
            raise ValueError ("unknown/unsupported SAT solver: "+self.SAT_solver)

    def run_open_wbo_solver (self):
        #return self.run_plingeling_or_open_wbo(["open-wbo", "-algorithm=5", "-cpu-lim=5", self.CNF_fname])
        return self.run_plingeling_or_open_wbo(["open-wbo", "-algorithm=5", self.CNF_fname])

    def run_maxsat_solver (self):
        return self.run_open_wbo_solver()

    def run_picomus (self):
        tmp_fname="tmp.out"
        err_fname="tmp.err"
        logfile=open(tmp_fname, "w")
        errfile=open(err_fname, "w")
        try:
            child = subprocess.Popen(["picomus", self.CNF_fname], stdout=logfile, stderr=errfile)
        except FileNotFoundError:
            print ("can't run picomus")
            exit(0)

        child.wait()
        logfile.flush()
        logfile.close()
        errfile.flush()
        errfile.close()

        errfile=open(err_fname, "r")
        line = errfile.readline()
        if len(line)>0:
            print ("error log:", line)
            exit(0)
        errfile.close()

        tmp=[]
        logfile=open(tmp_fname, "r")
        while True:
            line = logfile.readline()
            #print ("line=", line)
            if line.startswith("v "):
                tmp.append (int(line[2:].rstrip()))
            elif line=='':
                break
            else:
                pass

        logfile.close()
        os.remove(tmp_fname)
        os.remove(err_fname)
        return tmp

    def get_MUS_vars (self):
        self.write_CNF()
        clauses=self.run_picomus ()
        #print ("get_MUS_vars(), clauses:", clauses)
        vars=set()
        for clause in clauses:
            if clause==0:
                break
            s=self.CNF_clauses[clause-1]
            #print ("get_MUS_vars(), s:", s)
            for tmp in s:
                vars.add (abs(int(tmp)))
        #print ("get_MUS_vars(), vars:", vars)
        return clauses, sorted(list(vars))

    def write_CNF(self):
        VARS_TOTAL=self.last_var-1
        f=open(self.CNF_fname, "w")
        if self.maxsat==False:
            f.write ("p cnf "+str(VARS_TOTAL)+" "+str(self.clauses_total)+"\n")
        else:
            f.write ("p wcnf "+str(VARS_TOTAL)+" "+str(self.clauses_total)+" "+str(self.HARD_CLAUSE)+"\n")

        for i in range(self.CNF_next_idx):
            if i in self.CNF_clauses:
                line=" ".join(map(str, self.CNF_clauses[i]+[0]))+"\n"
            elif i in self.CNF_comments:
                line="c "+self.CNF_comments[i]+"\n"
            f.write(line)

        f.close()
        if self.verbosity>0:
            print ("write_CNF() clauses=%d" % self.clauses_total)
        self.CNF_written=True

    def create_var(self):
        self.last_var=self.last_var+1
        return self.last_var-1

    # FIXME: get rid of it!
    def neg(self, v:int) -> int:
        #print ("neg:", v)
        #if v==None:
        #    raise ValueError
        if v==0:
            raise ValueError
        return -v

    def neg_if(self, cond:bool, var:int) -> int:
        if cond:
            return self.neg(var)
        else:
            return var

    def BV_neg(self, lst:List[int]) -> List[int]:
        return [self.neg(l) for l in lst]

    # to be tested...
    def add_comment(self, comment:str):
        self.CNF_comments[self.CNF_next_idx]=comment
        self.CNF_next_idx=self.CNF_next_idx+1

    def add_clause(self, cls:List[int]):
        if self.verbose>=2:
            print ("add_clause", cls)
        self.clauses_total=self.clauses_total+1

        if self.SAT_solver=="libpicosat":

           #for v in cls:
           #    self.libpicosat.add(v)
           #self.libpicosat.add(0)

           self.libpicosat.add_lits(cls+[0])
           return

        if self.maxsat==False:
            self.CNF_clauses[self.CNF_next_idx]=cls
        else:
            self.CNF_clauses[self.CNF_next_idx]=[self.HARD_CLAUSE]+cls
        self.CNF_next_idx=self.CNF_next_idx+1

        #if (self.clauses_total % 1000000)==0:
        #    print "(hearbeat) add_clause(). clauses_total=", self.clauses_total

    def add_clauses(self, clauses:List[List[int]]):
        for cls in clauses:
            self.add_clause(cls)

    def add_soft_clause(self, cls:List[int], weight:int):
        assert self.maxsat==True
        assert weight>0
        assert type(cls)==list
        #print ("cls=", cls)
        #print ("weight=", weight)
        self.clauses_total=self.clauses_total+1
        self.CNF_clauses[self.CNF_next_idx]=[weight]+cls
        self.CNF_next_idx=self.CNF_next_idx+1

    def AND_Tseitin(self, v1:int, v2:int, out:int):
        self.add_clause([self.neg(v1), self.neg(v2), out])
        self.add_clause([v1, self.neg(out)])
        self.add_clause([v2, self.neg(out)])
    
    def AND(self, v1:int, v2:int) -> int:
        if (v1, v2) in self.AND_already_generated:
            return self.AND_already_generated[(v1,v2)]

        out=self.create_var()
        self.AND_Tseitin(v1, v2, out)

        self.AND_already_generated[(v1,v2)]=out
        return out
    
    def AND_list(self, l:List[int]) -> int:
        assert(len(l)>=1)

        # this is correct!
        if len(l)==1:
            return l[0]

        if len(l)==2:
            return self.AND(l[0], l[1])

        if tuple(l) in self.AND_list_already_generated:
            return self.AND_list_already_generated[tuple(l)]

        out=self.AND(l[0], self.AND_list(l[1:]))
        self.AND_list_already_generated[tuple(l)]=out
        return out

    # AKA 'not both'
    def NAND(self, v1:int, v2:int) -> int:
        if self.verbose>=1:
            print ("NAND", v1, v2)
        if (v1, v2) in self.NAND_already_generated:
            return self.NAND_already_generated[(v1,v2)]

        out=self.NOT(self.AND(v1, v2))

        self.NAND_already_generated[(v1,v2)]=out
        return out

    def BV_AND(self, x:List[int], y:List[int]) -> List[int]:
        rt=[]
        for pair in zip(x, y):
            rt.append(self.AND(pair[0],pair[1]))
        return rt
    
    # as in Tseitin transformations.
    # N.B.: previously called "OR"
    def OR_list(self, vals:List[int]) -> int:
        if self.verbose>=1:
            print ("OR_list", vals)
        if tuple(vals) in self.OR_list_already_generated:
            rt=self.OR_list_already_generated[tuple(vals)]
            if self.verbose>=1:
                print ("OR_list -> cached", rt)
            return rt

        #print (vals)
        if len(vals)==0:
            raise ValueError ("OR_list() requires list of >=1 vars")
        if len(vals)==1:
            return vals[0] # this is correct!
        out=self.create_var()
        self.add_clause(vals+[self.neg(out)])
        for v in vals:
            self.add_clause([self.neg(v), out])

        self.OR_list_already_generated[tuple(vals)]=out
        return out

    def OR_always(self, vals:List[int]):
        self.add_clause(vals)
    
    def alloc_BV(self, n:int) -> List[int]:
        return [self.create_var() for i in range(n)]
    
    def fix_soft(self, var:int, b:bool, weight:int):
        if b:
            self.add_soft_clause([var], weight)
        else:
            self.add_soft_clause([self.neg(var)], weight)

    def fix_soft_always_true(self, var:int, weight:int):
        self.fix_soft(var, True, weight)

    def fix_soft_always_true_all_bits_in_BV(self, BV:List[int], weight:int):
        for b in BV:
            self.fix_soft_always_true(b, weight)
    
    def fix(self, var:int, b:bool):
        if self.verbose>=1:
            print ("fix", var, b)
        if b:
            self.add_clause([var])
        else:
            self.add_clause([self.neg(var)])

    def fix_always_false(self, var:int):
        self.fix(var, False)

    def fix_always_true(self, var:int):
        self.fix(var, True)
    
    # BV is a list of True/False
    def fix_BV(self, _vars:List[int], BV:List[bool]):
        #print _vars, BV
        assert len(_vars)==len(BV)
        for var, _bool in zip(_vars, BV):
            self.fix (var, _bool)

    # BV is a list of True/False
    def fix_BV_soft(self, _vars:List[int], BV:List[bool], weight:int):
        assert len(_vars)==len(BV)
        for var, _bool in zip(_vars, BV):
            self.fix_soft (var, _bool, weight)

    def get_var_from_solution(self, var:int) -> int:
        # 1 if var is present in solution, 0 if present in negated form:
        if self.solution[int(var)]:
            return 1
        else:
            return 0
        raise ValueError ("incorrect var number: "+str(var))
    
    def get_BV_from_solution(self, BV:List[int]) -> List[int]:
        return [self.get_var_from_solution(var) for var in BV]
    
    def solve(self) -> bool:
        if self.CNF_initialized:
            self.write_CNF()
        if self.maxsat:
            self.solution=self.run_maxsat_solver()
        else:
            self.solution=self.run_sat_solver()
        if self.remove_CNF_file and os.path.exists(self.CNF_fname):
            os.remove(self.CNF_fname)
        if self.solution==None:
            return False
        else:
            return True

    def NOT(self, x:int) -> int:
        if self.verbose>=1:
            print ("NOT", x)
        if x in self.NOT_already_generated:
            return self.NOT_already_generated[x]

        rt=self.create_var()
        self.add_clause([self.neg(rt), self.neg(x)])
        self.add_clause([rt, x])

        self.NOT_already_generated[x]=rt
        return rt

    def BV_NOT(self, x:List[int]) -> List[int]:
        rt=[]
        for b in x:
            rt.append(self.NOT(b))
        return rt
    
    def XOR(self, x:int, y:int) -> int:
        if self.verbose>=1:
            print ("XOR", x, y)
        if (x, y) in self.XOR_already_generated:
            return self.XOR_already_generated[(x,y)]

        rt=self.create_var()
        self.add_clause([self.neg(x), self.neg(y), self.neg(rt)])
        self.add_clause([x, y, self.neg(rt)])
        self.add_clause([x, self.neg(y), rt])
        self.add_clause([self.neg(x), y, rt])

        self.XOR_already_generated[(x,y)]=rt
        return rt
    
    def BV_XOR(self, x:List[int], y:List[int]) -> List[int]:
        rt=[]
        for pair in zip(x,y):
            rt.append(self.XOR(pair[0], pair[1]))
        return rt

    def BV_XOR_list(self, l:List[List[int]]) -> List[int]:
        assert(len(l)>=2)
        if len(l)==2:
            return self.BV_XOR(l[0], l[1])
        return self.BV_XOR(l[0], self.BV_XOR_list(l[1:]))
    
    def EQ(self, x:int, y:int) -> int:
        if self.verbose>=1:
            print ("EQ", x, y)
        if (x, y) in self.EQ_already_generated:
            return self.EQ_already_generated[(x,y)]

        out=self.NOT(self.XOR(x,y))

        self.EQ_already_generated[x]=out
        return out
    
    def NEQ(self, x:int, y:int) -> int:
        return self.XOR(x,y)

    # p => q
    def IMPLY(self, p:int, q:int) -> int:
        if self.verbose>=1:
            print ("IMPLY", p, q)
        if (p, q) in self.IMPLY_already_generated:
            return self.IMPLY_already_generated[(p, q)]

        out=self.OR_list([self.NOT(p), q])
        self.IMPLY_already_generated[(p, q)]=out
        return out

    def IMPLY_always(self, p:int, q:int) -> int:
        return self.fix(self.IMPLY(p, q), True)

    # naive/pairwise/quadratic encoding
    def AtMost1_pairwise(self, lst:List[int]):
        for pair in itertools.combinations(lst, r=2):
            self.add_clause([self.neg(pair[0]), self.neg(pair[1])])

    # "commander" (?) encoding
    def AtMost1_commander(self, lst:List[int]) -> int:
        parts=my_utils.partition(lst, 5)
        c=[] # type: List[int]
        for part in parts:
            if len(part)<10:
                self.AtMost1_pairwise(part)
                c.append(self.OR_list(part))
            else:
                c.append(self.AtMost1_commander(part))
        self.AtMost1_pairwise(c)
        return self.OR_list(c)

    def AtMost1(self, lst:List[int]):
        if len(lst)<=10:
            self.AtMost1_pairwise(lst)
        else:
            self.AtMost1_commander(lst)

    # previously named POPCNT1 
    # make one-hot (AKA unitary) variable
    def make_one_hot(self, lst:List[int]):
        self.AtMost1(lst)
        self.OR_always(lst)

    def neg_nth_elem_in_lst(self, lst:List[int], n:int) -> List[int]:
        rt=[] # type: List[int]
        assert n<len(lst)
        for i in range(len(lst)):
            if i==n:
                rt.append(self.neg(lst[i]))
            else:
                rt.append(lst[i])
        return rt

    # sum of variables must not be equal to 1, but can be 0
    def SumIsNot1(self, lst:List[int]):
        for i in range(len(lst)):
            self.add_clause(self.neg_nth_elem_in_lst(lst, i))
    
    # Hamming distance between two bitvectors is 1
    # i.e., two bitvectors differ in only one bit.
    def hamming1(self, l1:List[int], l2:List[int]):
        self.add_comment("hamming1")
        assert len(l1)==len(l2)
        XORed=self.BV_XOR(l1, l2)
        self.make_one_hot(XORed)

    def fix_EQ(self, v1:int, v2:int):
        self.add_clause([self.neg(v1), v2])
        self.add_clause([v1, self.neg(v2)])
    
    # bitvectors must be equal to each other.
    def fix_BV_EQ(self, l1:List[int], l2:List[int]):
        #print len(l1), len(l2)
        assert len(l1)==len(l2)
        self.add_comment("fix_BV_EQ")
        for p in zip(l1, l2):
            self.fix_EQ(p[0], p[1])

    def BV_EQ(self, l1:List[int], l2:List[int]) -> int:
        #print len(l1), len(l2)
        assert len(l1)==len(l2)
        self.add_comment("BV_EQ")
        t=[]
        for p in zip(l1, l2):
            t.append(self.NOT(self.EQ(p[0], p[1])))
        return self.NOT(self.OR_list(t))
    
    # bitvectors must be different.
    def fix_BV_NEQ(self, l1:List[int], l2:List[int]):
        #print len(l1), len(l2)
        assert len(l1)==len(l2)
        self.add_comment("fix_BV_NEQ")
        t=[self.XOR(l1[i], l2[i]) for i in range(len(l1))]
        self.add_clause(t)

    # full-adder, as found by Mathematica using truth table:
    def FA (self, a:int, b:int, cin:int) -> Tuple[int, int]:
        s=self.create_var()
        cout=self.create_var()

        self.add_clause([self.neg(a), self.neg(b), self.neg(cin), s])
        self.add_clause([self.neg(a), self.neg(b), cout])
        self.add_clause([self.neg(a), self.neg(cin), cout])
        self.add_clause([self.neg(a), cout, s])
        self.add_clause([a, b, cin, self.neg(s)])
        self.add_clause([a, b, self.neg(cout)])
        self.add_clause([a, cin, self.neg(cout)])
        self.add_clause([a, self.neg(cout), self.neg(s)])
        self.add_clause([self.neg(b), self.neg(cin), cout])
        self.add_clause([self.neg(b), cout, s])
        self.add_clause([b, cin, self.neg(cout)])
        self.add_clause([b, self.neg(cout), self.neg(s)])
        self.add_clause([self.neg(cin), cout, s])
        self.add_clause([cin, self.neg(cout), self.neg(s)])

        return s, cout

    # bit order: [MSB..LSB]
    # n-bit adder:
    def adder(self, X:List[int], Y:List[int]) -> Tuple[List[int], int]:
        assert len(X)==len(Y)
        # first full-adder could be half-adder
        # start with lowest bits:
        inputs=my_utils.rvr(list(zip(X,Y)))
        carry=self.const_false
        sums=[]
        for pair in inputs:
            # "carry" variable is replaced at each iteration.
            # so it is used in the each FA() call from the previous FA() call.
            s, carry = self.FA(pair[0], pair[1], carry)
            sums.append(s)
        return my_utils.rvr(sums), carry

    # bit is 0 or 1.
    # i.e., if it's 0, output is 0 (all bits)
    # if it's 1, output=input
    def mult_by_bit(self, X:List[int], bit:int) -> List[int]:
        return [self.AND(i, bit) for i in X]

    # bit order: [MSB..LSB]
    # build multiplier using adders and mult_by_bit blocks:
    def multiplier(self, X:List[int], Y:List[int]) -> List[int]:
        assert len(X)==len(Y)
        out=[]
        #initial:
        prev=[self.const_false]*len(X)
        # first adder can be skipped, but I left thing "as is" to make it simpler
        for Y_bit in my_utils.rvr(Y):
            s, carry = self.adder(self.mult_by_bit(X, Y_bit), prev)
            out.append(s[-1])
            prev=[carry] + s[:-1]
    
        return prev + my_utils.rvr(out)

    def NEG(self, x:List[int]) -> List[int]:
        # invert all bits
        tmp=self.BV_NOT(x)
        # add 1
        one=self.alloc_BV(len(tmp))
        self.fix_BV(one, n_to_BV(1, len(tmp)))
        return self.adder(tmp, one)[0]

    # untested (?)
    def shift_left (self, x:List[int], cnt:int) -> List[int]:
        return x[cnt:]+[self.const_false]*cnt

    def shift_left_1 (self, x:List[int]) -> List[int]:
        return x[1:]+[self.const_false] 

    def shift_right (self, x:List[int], cnt:int) -> List[int]:
        return [self.const_false]*cnt+x[cnt:]

    def shift_right_1 (self, x:List[int]) -> List[int]:
        return [self.const_false]+x[:-1]

    def create_MUX(self, ins:List[int], sels:List[int]) -> int:
        assert 2**len(sels)==len(ins)
        x=self.create_var()
        for sel in range(len(ins)): # for example, 32 for 5-bit selector
            tmp=[self.neg_if((sel>>i)&1==1, sels[i]) for i in range(len(sels))] # 5 for 5-bit selector
   
            self.add_clause([self.neg(ins[sel])] + tmp + [x])
            self.add_clause([ins[sel]] + tmp + [self.neg(x)])
        return x
    
    # for 1-bit sel
    # ins=[[outputs for sel==0], [outputs for sel==1]]
    def create_wide_MUX (self, ins:List[List[int]], sels:List[int]) -> List[int]:
        out=[]
        for i in range(len(ins[0])):
            inputs=[x[i] for x in ins]
            out.append(self.create_MUX(inputs, sels))
        return out

    # untested:
    def ITE(self, s:int, f:int, t:int) -> int:
        if (s, f, t) in self.ITE_already_generated:
            return self.ITE_already_generated[(s,f,t)]

        if s==0:
            raise ValueError
        if f==0:
            raise ValueError
        if t==0:
            raise ValueError
        x=self.create_var()
        if x==0:
            raise ValueError
      
        # as found by my util 
        self.add_clause([self.neg(s),self.neg(t),x])
        self.add_clause([self.neg(s),t,self.neg(x)])
        self.add_clause([s,self.neg(f),x])
        self.add_clause([s,f,self.neg(x)])

        self.ITE_already_generated[(s,f,t)]=x
        return x

    def subtractor(self, minuend:List[int], subtrahend:List[int]) -> Tuple[List[int], int]:
        # same as adder(), buf: 1) subtrahend is inverted; 2) input carry-in bit is 1
        X=minuend
        Y=self.BV_NOT(subtrahend)

        inputs=my_utils.rvr(list(zip(X,Y)))
        carry=self.const_true
        sums=[]
        for pair in inputs:
            # "carry" variable is replaced at each iteration.
            # so it is used in the each FA() call from the previous FA() call.
            st, carry = self.FA(pair[0], pair[1], carry)
            sums.append(st)

        return my_utils.rvr(sums), carry

    # 0 if a<b
    # 1 if a>=b
    def comparator_GE(self, a:List[int], b:List[int]) -> int:
        tmp, carry = self.subtractor(a, b)
        return carry

    def div_blk(self, enable, divident:List[int], divisor:List[int]) -> Tuple[List[int], int]:
        assert len(divident)==len(divisor)

        diff, _ = self.subtractor(minuend=divident, subtrahend=divisor)

        cmp_res = self.AND(enable, self.comparator_GE(divident, divisor))

        out=self.alloc_BV(len(divident)) # is it used?!

        return self.create_wide_MUX([divident, diff], [cmp_res]), cmp_res

    def divider(self, divident:List[int], divisor:List[int]) -> Tuple[List[int], List[int]]:
        assert len(divident)==len(divisor)
        BITS=len(divisor)

        wide_divisor=self.shift_left([self.const_false]*BITS+divisor, BITS-1)
        quotient=[]
        for b in range(BITS):
            enable=self.NOT(self.OR_list(wide_divisor[:BITS]))
            divident, q_bit=self.div_blk(enable, divident, wide_divisor[BITS:])
            quotient.append(q_bit)
            wide_divisor=self.shift_right_1(wide_divisor)

        # remainder is left in divident:
        return quotient, divident

    def add_negated_solution_as_clause (self):
        negated_solution=[]
        for v in range(1, self.last_var):
            negated_solution.append(self.neg_if(self.get_var_from_solution(v), v))
        self.add_clause(negated_solution)
        
    def fetch_next_solution(self):
        self.add_negated_solution_as_clause()
        return self.solve()

    def count_solutions(self):
        if self.solve()==False:
            return 0

        cnt=1
        while True:
            if self.fetch_next_solution()==False:
                break
            cnt=cnt+1
        return cnt

    def get_all_solutions(self):
        if self.solve()==False:
            return None
        rt=[self.solution]

        while True:
            if self.fetch_next_solution()==False:
                break
            rt.append(self.solution)
        return rt

    def BV_not_zero(self, bv:List[int]) -> int:
        return self.OR_list(bv)

    def BV_zero(self, bv:List[int]) -> int:
        return self.NOT(self.OR_list(bv))

    def get_val_from_solution(self, var:List[int]) -> int:
        return BV_to_number(self.get_BV_from_solution(var))

    def make_distinct_BVs (self, lst:List[List[int]]):
        assert type(lst)==list
        assert type(lst[0])==list
        for pair in itertools.combinations(lst, r=2):
            self.fix_BV_NEQ(pair[0], pair[1])

    def sort_unit(self, a:int, b:int):
        return self.OR_list([a,b]), self.AND(a,b)

    def sorting_network_make_ladder(self, lst:List[int]) -> List[int]:
        if len(lst)==2:
            return list(self.sort_unit(lst[0], lst[1]))
    
        tmp=self.sorting_network_make_ladder(lst[1:]) # lst without head
        first, second=self.sort_unit(lst[0], tmp[0])
        return [first, second] + tmp[1:]

    def sorting_network(self, lst:List[int]) -> List[int]:
        # simplest possible, bubble sort
    
        if len(lst)==2:
            return self.sorting_network_make_ladder(lst)

        tmp=self.sorting_network_make_ladder(lst)
        return self.sorting_network(tmp[:-1]) + [tmp[-1]]

# G=list of tuples. each tuple is edge between two vertices
# each number is vertex
# total=number of vertices
# return: list, color for each vertex
def find_2_coloring_of_graph (G:List[Tuple[int,int]], total:int) -> List[int]:
    #print "find_2_coloring_of_graph begin"
    s=SAT_lib(False)
    colors=[s.alloc_BV(2) for p in range(total)]

    for i in G:
        #s.add(colors[i[0]]!=colors[i[1]])
        s.fix_BV_NEQ(colors[i[0]], colors[i[1]])

    assert s.solve()
    # get solution and return it:
    #print "find_2_coloring_of_graph end"
    return [BV_to_number(s.get_BV_from_solution(colors[p])) for p in range(total)]

